/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_OTHR_DEBUG			FALSE
#define FE_OTHR_THREAD_VERBOSE	FALSE

#define FE_OTHR_ATOM_GRAIN		100		//* TODO param

#include "operate/operate.pmh"

using namespace fe;

OperatorThreaded::~OperatorThreaded(void)
{
	finish();
}

void OperatorThreaded::setFullRange(sp<SurfaceRange> a_spRange)
{
//	feLog("fine atoms:\n%s\n",a_spRange->dump().c_str());
	m_spFullRange=a_spRange->combineAtoms(FE_OTHR_ATOM_GRAIN);
//	feLog("full atoms:\n%s\n",m_spFullRange->dump().c_str());
}

void OperatorThreaded::setNonAtomicRange(I32 a_start,I32 a_end)
{
	m_spFullRange=sp<SurfaceRange>(new SurfaceRange());
	m_spFullRange->nonAtomic().add(a_start,a_end);
}

void OperatorThreaded::setAtomicRange(
		sp<SurfaceAccessibleI> a_spSurfaceAccessibleI,
		OperatorSurfaceCommon::AtomicChange a_atomicChange,
		String a_group)
{
	sp<SurfaceRange> spRange=a_spSurfaceAccessibleI->atomize(
			SurfaceAccessibleI::AtomicChange(a_atomicChange),a_group,
			8*catalog<I32>("Threads"));

//	feLog("OperatorThreaded::setAtomicRange\n%s\n",
//			spRange->dump().c_str());

	m_spFullRange=spRange->combineAtoms(FE_OTHR_ATOM_GRAIN);
}

void OperatorThreaded::finish(void)
{
	const BWORD waitForJobs=m_waitForJobs;
	m_waitForJobs=FALSE;

	for(U32 m=0;m<m_threadsAlive;m++)
	{
		m_spGang->post(sp<SurfaceRange>(NULL));
	}
	if(m_spGang->workers())
	{
		m_spGang->finish();
	}
	m_waitForJobs=waitForJobs;
	m_threadsAlive=0;
}

void OperatorThreaded::initialize(void)
{
	catalog<I32>("Threads")=1;
	catalog<I32>("Threads","min")=1;
	catalog<I32>("Threads","high")=16;
	catalog<bool>("Threads","joined")=true;
	catalog<String>("Threads","IO")="input output";
	catalog<String>("Threads","hint")=
			"Number of workers to process the data in parallel."
			"  The number of jobs may not be the same."
			"  When the thread count is set to one,"
			" no threads are spawned and the entire"
			" processing job is done at once with the existing thread.";

	catalog<bool>("AutoThread")=false;
	catalog<String>("AutoThread","label")="Auto";
	catalog<bool>("AutoThread","joined")=true;
	catalog<String>("AutoThread","IO")="input output";
	catalog<String>("AutoThread","hint")=
			"Automatically set the number of threads.";

	catalog<bool>("Paging")=false;
	catalog<String>("Paging","label")="Page";
	catalog<bool>("Paging","joined")=true;
	catalog<String>("Paging","hint")=
			"Respect paging even when it can be avoided."
			"  This is relevant for Houdini 12.5+."
			"  Houdini 11 and Maya are never paged."
			"  Houdini 12.0 and 12.1 are always paged.";

	catalog<bool>("StayAlive")=false;
	catalog<String>("StayAlive","label")="Stay Alive";
	catalog<String>("StayAlive","hint")=
			"Keep the threads running and looking for jobs even"
			" when this node isn't being processed."
			"  This may speed up quick nodes where launching threads is"
			" a large portion of the processing time,"
			" but this technique should be considered tempermental"
			" (crashes may become more likely).";

	m_spGang=new Gang< Worker,sp<SurfaceRange> >();
}

BWORD OperatorThreaded::accessOutput(sp<SurfaceAccessibleI>& a_rspAccessible,
	Record& a_rSignal,Message a_message)
{
//	adjustThreads();

	const BWORD result=OperatorSurfaceCommon::accessOutput(
			a_rspAccessible,a_rSignal,a_message);

	if(result && a_rspAccessible.isValid())
	{
		a_rspAccessible->setPaging(catalog<bool>("Paging"));
	}

	return result;
}

void OperatorThreaded::adjustThreads(void)
{
	I32& rThreads=catalog<I32>("Threads");

	if(catalog<bool>("AutoThread"))
	{
		rThreads=Thread::hardwareConcurrency();
	}

	setThreading((rThreads==1)? e_singleThread: e_multiThread);
}

void OperatorThreaded::handle(Record& a_rSignal)
{
#if FE_OTHR_DEBUG
	feLog("OperatorThreaded::handle \"%s\"\n",name().c_str());
#endif

	adjustThreads();

	I32& rThreads=catalog<I32>("Threads");
	const bool stayAlive=(catalog<bool>("StayAlive") && rThreads>1);
	m_waitForJobs=stayAlive;

	if(stayAlive)
	{
		catalog<String>("warning")+="using Stay Alive is dangerous;";
	}

	if(m_threadsAlive &&
			(!stayAlive || (stayAlive && m_threadsAlive!=rThreads)))
	{
#if FE_OTHR_THREAD_VERBOSE
		feLog("Pre-finish stayAlive %d threadsAlive %d\n",
				stayAlive,m_threadsAlive);
#endif
		finish();
	}

	BWORD runDirect=(rThreads==1);
	if(!runDirect && catalogOrDefault<I32>("Refinement",0))
	{
		catalog<String>("warning")+=
				"bypassing multi-threading while Refinement is on;";
		setThreading(e_singleThread);
		runDirect=TRUE;
	}

	//* for a single thread, run directly
	if(runDirect)
	{
#if FE_OTHR_THREAD_VERBOSE
		feLog("OperatorThreaded::handle single thread %s\n",
				m_spFullRange->brief().c_str());
#endif

		run(-1,m_spFullRange);
		return;
	}

	//* TODO aggregate atomics when there is a whole lot of them

	const U32 atomicCount=m_spFullRange->atomicCount();
	for(U32 atomicIndex=0;atomicIndex<atomicCount;atomicIndex++)
	{
		const SurfaceRange::MultiSpan& rAtomic=
				m_spFullRange->atomic(atomicIndex);
		sp<SurfaceRange> spSubRange(new SurfaceRange());
		spSubRange->addAtomic()=rAtomic;

#if FE_OTHR_THREAD_VERBOSE
		feLog("Post atomic %s\n",spSubRange->brief().c_str());
#endif
		m_spGang->post(spSubRange);
	}

	const SurfaceRange::MultiSpan& rNonAtomic=m_spFullRange->nonAtomic();
	const U32 spanCount=rNonAtomic.spanCount();

	for(U32 spanIndex=0;spanIndex<spanCount;spanIndex++)
	{
		const SurfaceRange::Span& rSpan=rNonAtomic.span(spanIndex);
		const U32 count=rSpan[1]-rSpan[0]+1;

//		const U32 jobs=8*rThreads;	//* TODO param?
		const U32 jobs=rThreads;

		const U32 each=(count>jobs)? (count+jobs-1)/jobs: 1;

		for(U32 m=0;m<count;m+=each)
		{
			const I32 start=rSpan[0]+m;
			I32 end=start+each-1;
			if(end>rSpan[1])
			{
				end=rSpan[1];
			}

#if FE_OTHR_THREAD_VERBOSE
			feLog("Post non-atomic [%d,%d]\n",start,end);
#endif
			if(end>=start)
			{
				sp<SurfaceRange> spSubRange(new SurfaceRange());
				spSubRange->nonAtomic().add(start,end);

				m_spGang->post(spSubRange);
			}
		}
	}

	if(!stayAlive || !m_threadsAlive)
	{
#if FE_OTHR_THREAD_VERBOSE
		feLog("Start %d threads\n",rThreads);
#endif
		if(!m_spGang->start(rThreads,sp<OperatorThreaded>(this)))
		{
			feLog("OperatorThreaded::handle \"%s\" threading failed"
					" -> switching to single thread\n",name().c_str());

			m_threadsAlive=0;
			catalog<bool>("AutoThread")=FALSE;
			rThreads=1;
			setThreading(e_singleThread);
			run(-1,m_spFullRange);
			return;
		}
		if(stayAlive)
		{
			m_threadsAlive=rThreads;
		}
	}

	if(stayAlive)
	{
#if FE_OTHR_THREAD_VERBOSE
		feLog("Post-sync\n");
#endif
		for(U32 m=0;m<m_threadsAlive;m++)
		{
			m_spGang->post(sp<SurfaceRange>(NULL));
		}

		m_spGang->synchronize(TRUE);
	}
	else
	{
#if FE_OTHR_THREAD_VERBOSE
		feLog("Post-finish\n");
#endif
		finish();
	}

	setThreading(e_singleThread);

	if(atomicCount)
	{
		const SurfaceRange::MultiSpan& rPostAtomic=
				m_spFullRange->postAtomic();
		sp<SurfaceRange> spSubRange(new SurfaceRange());
		spSubRange->postAtomic()=rPostAtomic;

#if FE_OTHR_THREAD_VERBOSE
		feLog("Run post-Atomic %s\n",spSubRange->brief().c_str());
#endif
		run(-1,spSubRange);
	}

	m_spGang->resolveException();

#if FE_OTHR_DEBUG
	feLog("OperatorThreaded::handle done\n");
#endif
}

OperatorThreaded::Worker::Worker(U32 a_id,
	sp< JobQueue< sp<SurfaceRange> > > a_spJobQueue):
	m_id(a_id),
	m_spJobQueue(a_spJobQueue)
{
#if FE_OTHR_THREAD_VERBOSE
	feLog("OperatorThreaded::Worker::Worker %p %p id %d\n",
			this,(Thread::Functor*)this,m_id);
#endif
}

OperatorThreaded::Worker::~Worker(void)
{
#if FE_OTHR_THREAD_VERBOSE
	feLog("OperatorThreaded::Worker::~Worker %p %p id %d\n",
			this,(Thread::Functor*)this,m_id);
#endif
}

void OperatorThreaded::Worker::operate(void)
{
#if FE_OTHR_THREAD_VERBOSE
	feLog("Starting thread %d\n",m_id);
#endif

	//* TODO confirm this couldn't change while reading
	const BWORD& rWait=m_spOperatorThreaded->m_waitForJobs;

	BWORD done=FALSE;
	while(!done)
	{
		sp<SurfaceRange> spRange;

		if(rWait)
		{
#if FE_OTHR_THREAD_VERBOSE
			feLog("Thread %d waiting\n",m_id);
#endif
			m_spJobQueue->waitForJob(spRange);
		}
		else
		{
			done=!m_spJobQueue->take(spRange);
		}

#if FE_OTHR_THREAD_VERBOSE
		feLog("Thread %d Job %s\n",m_id,
				spRange.isValid()? spRange->brief().c_str(): "NULL");
#endif

		if(spRange.isNull() || spRange->empty())
		{
			if(rWait)
			{
#if FE_OTHR_THREAD_VERBOSE
				feLog("Sync thread %d supervised\n",m_id);
#endif
				m_spJobQueue->synchronize(TRUE);
			}
			else
			{
				done=TRUE;
			}
		}
		else if(!done)
		{
			//* propagate exceptions to parent thread
			try
			{
				m_spOperatorThreaded->run(m_id,spRange);
			}
			catch(const Exception &e)
			{
#if FE_OTHR_THREAD_VERBOSE
				feLog("OperatorThreaded::Worker::operator()"
						" caught fe::Exception %s\n",c_print(e));
#endif
				m_spJobQueue->relayException(e);
			}
			catch(const std::exception& std_e)
			{
#if FE_OTHR_THREAD_VERBOSE
				feLog("OperatorThreaded::Worker::operator()"
						" caught std::exception %s\n",std_e.what());
#endif
				Exception e("stdlib",std_e.what());
				m_spJobQueue->relayException(e);
			}
#ifdef FE_BOOST_EXCEPTIONS
			catch(const boost::exception& boost_e)
			{
#if FE_OTHR_THREAD_VERBOSE
				feLog("OperatorThreaded::Worker::operator()"
						" caught boost::exception %s\n",
						boost::diagnostic_information(boost_e).c_str());
#endif
				Exception e("boost",
						boost::diagnostic_information(boost_e).c_str());
				m_spJobQueue->relayException(e);
			}
#endif
			catch(...)
			{
#if FE_OTHR_THREAD_VERBOSE
				feLog("OperatorThreaded::Worker::operator()"
						" caught unknown exception\n");
#endif
				Exception e;
				m_spJobQueue->relayException(e);
			}
		}
	}

	//* Do we need to sync if we join right afterwards?
/*
#if FE_OTHR_THREAD_VERBOSE
	feLog("Sync thread %d\n",m_id);
#endif

	const U32 leader=m_spJobQueue->synchronize();

#if FE_OTHR_THREAD_VERBOSE
	feLog("Sync'd thread %d leader=%d\n",m_id,leader);
#endif
*/
}
