/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define	FE_TERM_DEBUG	FALSE

#include "terminal.pmh"

namespace fe
{

/***
	incoming commands:
		<result> = set "callback" <pointer>
		<result> = set <name> <variable> <type> <value>
		<result> = set <surf> model <type> <pointer>		TODO

		<result> = manage <dl>
		<result> = create <implementation> [name|surf]
		<result> = signal <name> [surfOut] [surfIn0] ...
		<result> = release <name>

		<result> = load <surf> <filename>
		<result> = save <surf> <filename>

	callback requests:
		<value> = get <name> <key>

	issues:
		need an abstracted get() to be used by the HandlerI instance

	TODO:
		jsoncpp for geo files
*/

void Terminal::initialize(void)
{
	m_spScope=registry()->master()->catalog()->catalogComponent(
			"Scope","SimScope");
}

String Terminal::execute(String a_commandline)
{
#if FE_TERM_DEBUG
	feLog("Terminal::execute \"%s\"\n",a_commandline.c_str());
#endif

	const String command=a_commandline.parse();

	if(command=="set")
	{
		const String target=a_commandline.parse();
		if(target=="callback")
		{
			String pointer=a_commandline.parse();

			sscanf(pointer.c_str(),"%p",&m_fpCallback);
			return "changed callback";
		}
		if(m_nodeMap.find(target)!=m_nodeMap.end())
		{
			sp<Catalog> spCatalog=m_nodeMap[target].catalog();

			const String variable=a_commandline.parse();
			if(!variable.empty())
			{
				const String type=a_commandline.parse();
				if(!type.empty())
				{
					const String value=a_commandline.parse();
					if(!value.empty())
					{
						if(type=="string")
						{
							spCatalog->catalog<String>(variable)=value;
						}
						else if(type=="boolean")
						{
							spCatalog->catalog<bool>(variable)=value.integer();
						}
						else if(type=="integer")
						{
							spCatalog->catalog<I32>(variable)=value.integer();
						}
						else if(type=="real")
						{
							spCatalog->catalog<Real>(variable)=value.real();
						}
						else if(type=="vector3")
						{
							Real x=Real(0);
							Real y=Real(0);
							Real z=Real(0);
							sscanf(value.c_str(),"%f%f%f",&x,&y,&z);
							spCatalog->catalog<SpatialVector>(variable)=
									SpatialVector(x,y,z);
						}
						else
						{
							return "error: unknown type";
						}
						return "changed catalog";
					}
					return "error: missing value";
				}
				return "error: missing variable";
			}
			return "error: missing type";
		}
		return "error: unknown target";
	}

	if(command=="manage")
	{
		const String name=a_commandline.parse();

		Result result=registry()->manage(name);
		if(successful(result))
		{
			return "managed";
		}
		return "error: unknown library";
	}

	if(command=="create")
	{
		const String implementation=a_commandline.parse();
		if(implementation.empty())
		{
			return "error: missing implementation";
		}

		const String name=a_commandline.parse();
		if(name.empty())
		{
			return "error: missing name";
		}

		m_nodeMap[name].setupOperator(m_spScope,implementation,name);

		return "created";
	}

	if(command=="release")
	{
		const String name=a_commandline.parse();

		if(m_nodeMap.find(name)!=m_nodeMap.end())
		{
			m_nodeMap.erase(name);
			return "released node";
		}
		else if(m_surfaceMap.find(name)!=m_surfaceMap.end())
		{
			m_surfaceMap.erase(name);
			return "released surface";
		}

		return "error: unknown target";
	}

	if(command=="signal")
	{
		const String instance=a_commandline.parse();

		sp<Catalog> spCatalog=m_nodeMap[instance].catalog();

		String outputName;
		String outputKey;

		I32 surfaceIndex= -1;
		while(TRUE)
		{
			sp<SurfaceAccessibleI> spSurfaceAccessibleI;

			const String surfaceName=a_commandline.parse();
			if(surfaceIndex<0)
			{
				outputName=surfaceName;
			}
			if(surfaceName.empty())
			{
				if(a_commandline.empty())
				{
					break;
				}
			}
			else if(m_surfaceMap.find(surfaceName)!=m_surfaceMap.end())
			{
				spSurfaceAccessibleI=
						m_surfaceMap[surfaceName].m_spSurfaceAccessibleI;
			}
			else if(surfaceIndex>=0)
			{
				return "error: unknown surface \"" + surfaceName + "\"";
			}

			if(surfaceIndex>=0 && spSurfaceAccessibleI.isNull())
			{
				return "error: invalid surface \"" + surfaceName + "\"";
			}

			//* multiple inputs (named by index)

			String lookup="output:0";
			if(surfaceIndex>=0)
			{
				lookup.sPrintf("input:%d",surfaceIndex);
			}
			const String key=spCatalog->catalog<String>(lookup);

//			feLog("surface \"%s\" is really \"%s\"\n",
//					lookup.c_str(),key.c_str());

			spCatalog->catalog< sp<Component> >(key)=spSurfaceAccessibleI;

			//* copy first input to output
			if(surfaceIndex==0)
			{
				if(spSurfaceAccessibleI.isValid())
				{
					sp<SurfaceAccessibleI> spOutput(
							dynamic_cast<SurfaceAccessibleI*>(
							spSurfaceAccessibleI->clone()));
					spOutput->setName(outputName);
					spOutput->bind(m_spScope);

					spCatalog->catalog< sp<Component> >(outputKey)=spOutput;

					m_surfaceMap[outputName].m_spSurfaceAccessibleI=spOutput;
				}
			}
			else if(surfaceIndex<0)
			{
				outputKey=key;
			}

			surfaceIndex++;
		}

		const BWORD success=m_nodeMap[instance].cook();
		if(!success)
		{
			m_surfaceMap.erase(outputName);
		}

/*
		// TEMP make request directly
		if(!m_fpCallback)
		{
			return "error: no callback";
		}

		String request;
		request.sPrintf("get %s args",instance.c_str());

#if FE_TERM_DEBUG
		feLog("Terminal::execute request \"%s\"\n",request.c_str());
#endif

		const std::string response=m_fpCallback(request.c_str());

#if FE_TERM_DEBUG
		feLog("Terminal::execute response \"%s\"\n",response.c_str());
#endif
*/

		return success? "signaled": "error: signal failed";
	}

	if(command=="load")
	{
		const String surfaceName=a_commandline.parse();
		if(surfaceName.empty())
		{
			return "error: surface name missing";
		}

		const String filename=a_commandline.parse();
		if(filename.empty())
		{
			return "error: filename missing";
		}

		const String suffix=String(strrchr(filename.c_str(),'.')).prechop(".");

		Surface& rSurface=m_surfaceMap[surfaceName];
		sp<SurfaceAccessibleI>& rspSurfaceAccessibleI=
				rSurface.m_spSurfaceAccessibleI;

		sp<Registry> spRegistry=m_spScope->registry();
		rspSurfaceAccessibleI=
				spRegistry->create("SurfaceAccessibleI.*.*."+suffix);
		rspSurfaceAccessibleI->setName(surfaceName);
		rspSurfaceAccessibleI->bind(m_spScope);
		const BWORD success=rspSurfaceAccessibleI->load(filename);

		return success? "loaded": "error: load failed";
	}

	if(command=="save")
	{
		const String surfaceName=a_commandline.parse();
		if(surfaceName.empty())
		{
			return "error: surface name missing";
		}

		if(m_surfaceMap.find(surfaceName)==m_surfaceMap.end())
		{
			return "error: unknown surface";
		}

		const String filename=a_commandline.parse();
		if(filename.empty())
		{
			return "error: filename missing";
		}

		Surface& rSurface=m_surfaceMap[surfaceName];

		if(rSurface.m_spSurfaceAccessibleI.isNull())
		{
			return "error: null surface";
		}

		const BWORD success=rSurface.m_spSurfaceAccessibleI->save(filename);

		return success? "saved": "error: save failed";
	}

	return "error: unknown command";
}

} // namespace
